# maak/__init__.py
#
#

"""
    Verhaal der Gemaakten ...

"""

__version__ = 2

## =============
## BASIC IMPORTS 
## =============

import collections
import traceback
import threading
import getpass
import logging
import hashlib
import optparse
import _thread
import random
import socket
import string
import fcntl
import types
import errno
import queue
import uuid
import json
import time
import imp
import sys
import os
import re

## maak imports

from maak import drivers

## oolib imports

from oolib import Error, Base, Dispatcher, Plugins, cb, config, colors
from oolib.utils import *

## config

config.workdir = ".maak"

## ======= 
## defines 
## ======= 

basic_types= [str, int, float, bool, None]

## =========
## VARIABLES
## =========

homedir = os.path.expanduser("~")

# check wether ocontrib is available

if os.path.isdir("contrib"): sys.path.append("contrib")

## ===============
## OPTION HANDLING 
## ===============

## make_opts function

def make_opts():
    parser = optparse.OptionParser(usage='usage: %prog [options]', version=make_version())
    for option in options:
        type, default, dest, help = option[2:]
        if "store" in type:
            try: parser.add_option(option[0], option[1], action=type, default=default, dest=dest, help=help)
            except Exception as ex: logging.error("error: %s - option: %s" % (str(ex), option)) ; continue
        else:
            try: parser.add_option(option[0], option[1], type=type, default=default, dest=dest, help=help)
            except Exception as ex: logging.error("error: %s - option: %s" % (str(ex), option)) ; continue
    # return a (opts, args) pair
    return parser.parse_args()


def dispatch(target, event, cmnd, *args, **kwargs):
     logging.info("dispatch %s on %s" % (cmnd, get_cls(target)))
     try: func = target[cmnd]               
     except KeyError: logging.debug("no %s found" % cmnd) ; return
     try: res = func(event)
     except: error() ; res = None
     return res

## ==========
## EXCEPTIONS
## ==========

class NoEvent(Error): pass

class RemoteDisconnect(Error): pass

## ==========
## CORE STUFF
## ==========

## XMPP class

class XMPP(dict):

    def make_xmpp(self):
        import sleekxmpp
        from sleekxmpp.xmlstream.tostring import xml_escape
        res = dict(self)
        try: del res["from"]
        except: pass
        elem = self['element']
        main = "<%s" % self['element']
        for attribute in attributes[elem]:
            if attribute in res:
                if res[attribute]: main += " %s='%s'" % (attribute, xml_escape(stripbadchar(str(res[attribute]))))
                continue
        main += ">"
        if "xmlns" in res: main += "<x xmlns='%s'/>" % res["xmlns"] ; gotsub = True
        else: gotsub = False
        if 'html' in res:   
            if res['html']: 
                main += '<html xmlns="http://jabber.org/protocol/xhtml-im"><body xmlns="http://www.w3.org/1999/xhtml">%s</body></html>' % res['html']
                gotsub = True
        if 'txt' in res:     
            if res['txt']:
                txt = res['txt']   
                main += "<body>%s</body>" % stripbadchar(xml_escape(txt))
                gotsub = True
        for subelement in subelements[elem]:   
            if subelement == "body": continue  
            if subelement == "thread": continue
            try:
                data = res[subelement]
                if data:
                    try:
                        main += "<%s>%s</%s>" % (subelement, xml_escape(data), subelement)
                        gotsub = True
                    except AttributeError as ex: logging.warn("skipping %s" % subelement)
            except KeyError: pass
        if gotsub: main += "</%s>" % elem
        else: main = main[:-1] ; main += " />"
        return main

## Event class

class Event(Base): 

    def prepare(self):
        try: self.first, self.rest = self.txt.split(" ", 1)
        except: self.first = self.txt
        if self.rest: self.args = self.rest.split()
        if self.first and self.first[0] == ".": self.user_cmnd = self.first[1:]

    def handle(self, *args , **kwargs): pass

## Bot class

class Bot(Dispatcher): 

    def connect(self, *args, **kwargs): pass

    def exit(self, *args, **kwargs): pass

    def get_one(self): pass

    def read_some(self): pass

    def do_once(self, *args, **kwargs):
        event = args[0]
        logging.debug("got event %s" % str(event))
        event._target = self
        event._state = "once"
        self.put(event)
        return event

    def handle_once(self, *args, **kwargs):
        if args: event = args[0]
        else: event = self.get_one()
        logging.debug("got event %s" % str(event))
        event._target = self
        self.put(event)
        return event

    def run_forever(self, *args, **kwargs):
        logging.warn("starting %s" % get_clsname(self))
        self.wait()
        try: self.connect()
        except socket.gaierror: error() ; self._state = "error"
        else: self._state = "running"
        while self._state in ["running", "callback", "idle", "empty", "once"]:
             event = self.handle_once() ; event.wait()
        logging.debug("stopping %s (%s)" % (self.what, self._state))

## Fleet class

class Fleet(Base):

    def register(self, name, bot):
        if not self.bots: self.bots = []
        logging.warn("register %s in fleet" % name)
        self.bots.append(bot)

    def start(self, *args, **kwargs):
        event = Event()
        if args:
            config = args[0]
            event.txt = "."
            for arg in config.args:
                event.txt += "%s " % arg
        for bot in self.bots:
            try: _thread.start_new_thread(bot.run_forever, ())
            except: error()
            if event.txt: bot.put(event)
        if self.bots and "once" not in kwargs:
            logging.warn("bots in fleet: %s" % ", ".join([get_cls(bot) for bot in self.bots]))
            logging.warn("commands are: %s" % " ".join(cmnds.keys()))
            for bot in self.bots: bot.ready()
            while 1:
                try: time.sleep(1)
                except KeyboardInterrupt: break
                except Exception: error() ; break
        shutdown()

    def exit(self):
        for bot in self.bots:
            try: bot.exit()
            except AttributeError: continue
            except: error()

## Config class

class Config(Base): pass

## Commands class

class Commands(Base): pass

## =========
## GREETINGS
## =========

## make_version function

def make_version(): return "%sVerhaal der Gemaakten #%s\n%s%s" % (colors.YELLOW, __version__, time.ctime(time.time()), colors.ENDC)

## hello function

def hello(): print(make_version() + "\n")

## ==========
## start/stop
## ==========

## boot function

def boot():
    global config
    try: config.opts, config.args = make_opts()
    except SystemExit: os._exit(1)
    config.update(vars(config.opts))
    if config.doshell: hello()
    for arg in config.args:
        try: var, val = arg.split("=") ; config[var] = val ; continue
        except ValueError: pass
    if config.do_local: config.workdir = j(os.getcwd(), ".maak") 
    if not config.workdir: config.workdir = ".maak"
    make_dir(config.workdir)
    if not config.loglevel: config.loglevel = "error"
    from oolib.log import log_config
    log_config(config.loglevel)
    if config.loglevel:
        logging.warn("C O N F I G")
        logging.warn("")
        for line in config.show_items():
            logging.warn(line)
    if config.loglevel:
        logging.warn("")
        logging.warn("E G G S")
        logging.warn("")
        show_eggs()
        show_eggs("maak")
    if config.loglevel:
        logging.warn("")
        logging.warn("B O O T")
        logging.warn("")
    set_core()
    path, fn = os.path.split(os.path.abspath(__file__))
    plugins.load_plugs(path)
    logging.warn("")
    logging.warn("R E A D Y")
    logging.warn("")
    return config

## shutdown function

def shutdown():
    print("")
    logging.warn("shutdown has arrived")
    fleet.exit()
    os._exit(0)

## +++++++++++++++++++++
## BASIC RUNTIME OBJECTS
## +++++++++++++++++++++

cmnds = Commands()
tests = Commands()
fleet = Fleet()
config = Config()
plugins = Plugins()

## core holder

core = Base()

def set_core():
    core.register("cb", cb)
    core.register("cmnds", cmnds)
    core.register("tests", tests)
    core.register("fleet", fleet)
    core.register("config", config)
    core.register("plugins", plugins)

## THNX !!
